#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include <alsa/asoundlib.h>

#include "usb.h"

#define MIDI_ID_SIZE 10
#define MIDI_BUFFER_SIZE 128


int quit=0;


struct midi_device
{
 snd_rawmidi_t *input;
 snd_rawmidi_t *output;
};


struct buffer
{
 uint16_t size;
 uint16_t length;
 unsigned char *data;
};


/*
 * Signals
 */

void signalHandler(int signal, siginfo_t *info, void *object)
{
 switch(signal)
 {
  case SIGPIPE:
  case SIGHUP:
  {
   break;
  }
  default:
  {
   quit=1;
   break;
  }
 }
}

int initSignals()
{
 struct sigaction sa;
 sigemptyset(&sa.sa_mask);
 sa.sa_flags=0;
 sa.sa_sigaction=signalHandler;

 if(
    sigaction(SIGPIPE, &sa, 0)<0||
    sigaction(SIGCHLD, &sa, 0)<0||
    sigaction(SIGTERM, &sa, 0)<0||
    sigaction(SIGQUIT, &sa, 0)<0||
    sigaction(SIGABRT, &sa, 0)<0||
    sigaction(SIGSEGV, &sa, 0)<0||
    sigaction(SIGALRM, &sa, 0)<0||
    sigaction(SIGINT, &sa, 0)<0||
    sigaction(SIGHUP, &sa, 0)<0
 )
  return 1;

 return 0;
}


/*
 * Buffer
 */

int initBuffer(struct buffer *buffer)
{
 buffer->size=512;
 buffer->length=0;
 buffer->data=NULL;

 buffer->data=(unsigned char *)malloc(sizeof(unsigned char)*buffer->size); //allocate memory

 if(!buffer->data)
  return 1;

 memset(buffer->data, 0, buffer->size); //fill with zeros
 return 0;
}

int deinitBuffer(struct buffer *buffer)
{
 if(buffer->data)
  free(buffer->data);

 buffer->size=0;
 buffer->length=0;
 buffer->data=NULL;

 return 0;
}

void dumpBuffer(struct buffer *buffer)
{
 int c;

 printf("DATA: ");

 if(!buffer->size)
  printf("not initialized");
 else if(!buffer->length)
  printf("empty");
 else
 {
  for(c=0;c<buffer->length;c++)
   printf("%.2x ", buffer->data[c]);
 }

 printf("\n");
}

int copyBuffer(struct buffer *buffer, const unsigned char *data, uint16_t length)
{
 if(!buffer->data||!buffer->size)
 {
  printf("Buffer not initialized\n");
  return 1;
 }

 if(buffer->length) //overwrite
  memset(buffer->data, 0, buffer->size); //fill with zeros

 if(length>buffer->size)
 {
  printf("Data (%d) exceeds buffer size (%d), skipping last %d bytes\n", length, buffer->size, (length-buffer->size));
  memcpy(buffer->data, data, buffer->size);
  buffer->length=buffer->size;
 }
 else if(data&&length) //copy data
 {
  memcpy(buffer->data, data, length);
  buffer->length=length;
 }

 return 0;
}


/*
 * Roland
 */

int initRolandTD15(struct usb_device *dev)
{
 struct usb_transfer transfer;
 transfer.data=NULL; //ensure data isn't freed the first time


 /* open device */

 dev->vendor=0x0582; dev->product=0x0150; //Roland / TD-15

 if(usb_open(dev, 3)!=0)
  return 1;


 /* init device */

 usb_transfer_reset(&transfer, 18, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x100;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 9, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 167, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 64, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x100;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 18, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x100;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 255, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 255, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x300;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 255, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x302;
 transfer.index=0x0409;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 18, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x100;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 80, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 167, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x01;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x02;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x02;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x01;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 return 0;
}

int deinitRolandTD15(struct usb_device *dev)
{
 struct usb_transfer transfer;
 transfer.data=NULL; //ensure data isn't freed the first time


 /* deinit device */

 usb_transfer_reset(&transfer, 18, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x100;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 167, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_DEVICE;
 transfer.request=LIBUSB_REQUEST_SET_CONFIGURATION;
 transfer.value=0x01;
 transfer.index=0x00;

 if(usb_control(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  usb_close(dev);
  return 2;
 }


 /* close device */

 usb_transfer_reset(&transfer, 0, NULL);
 usb_close(dev);
 return 0;
}

int readRolandTD15(struct usb_device *dev, struct buffer *buffer)
{
 int c;

 struct usb_transfer transfer;
 transfer.data=NULL; //ensure data isn't freed the first time

 usb_transfer_reset(&transfer, 512, NULL);
 transfer.recipient=LIBUSB_RECIPIENT_ENDPOINT;
 transfer.endpoint=LIBUSB_ENDPOINT_IN;

 if(usb_bulk(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  return 3;
 }

 if(transfer.length) //data
 {
  printf("TD15 read  %d: ", transfer.length);
  for(c=0;c<transfer.length;c++)
   printf("%.2x ", transfer.data[c]);

  printf("\n");
 }

 if(buffer)
  copyBuffer(buffer, transfer.data, transfer.length);

 usb_transfer_reset(&transfer, 0, NULL);
 return 0;
}

int writeRolandTD15(struct usb_device *dev, struct buffer *buffer)
{
 int c;

 struct usb_transfer transfer;
 transfer.data=NULL; //ensure data isn't freed the first time

 if(!buffer||!buffer->length)
 {
  printf("Nothing to write to Roland TD-15\n");
  return 3;
 }

 if(buffer->length) //data
 {
  printf("TD15 write %d: ", buffer->length);
  for(c=0;c<buffer->length;c++)
   printf("%.2x ", buffer->data[c]);

  printf("\n");
 }

 usb_transfer_reset(&transfer, buffer->length, NULL);
 transfer.recipient=LIBUSB_RECIPIENT_OTHER;
 transfer.endpoint=LIBUSB_ENDPOINT_OUT;

 //copy data
 memcpy(transfer.data, buffer->data, buffer->length);
 transfer.length=buffer->length;

 if(usb_bulk(dev, &transfer)!=0)
 {
  usb_transfer_reset(&transfer, 0, NULL);
  return 3;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 return 0;
}


/*
 * MIDI
 */

int initMIDI(struct midi_device *dev, const char *devID)
{
 int r;
 char midiID[MIDI_ID_SIZE+3];

 dev->input=NULL;
 dev->output=NULL;

 if(strncmp("hw:", devID, 3)!=0) //no hw: prefix
  sprintf(midiID, "hw:%s", devID); //append prefix
 else //hw prefix found
  strncpy(midiID, devID, MIDI_ID_SIZE); //copy id

 if((r=snd_rawmidi_open(&dev->input, &dev->output, midiID, SND_RAWMIDI_NONBLOCK))<0)
 {
  printf("Can't open MIDI input: %s\n", snd_strerror(r));
  return 1;
 }

 return 0;
}

int deinitMIDI(struct midi_device *dev)
{
 if(dev->input)
 {
  snd_rawmidi_close(dev->input);
  dev->input=NULL;
 }

 if(dev->output)
 {
  snd_rawmidi_close(dev->output);
  dev->output=NULL;
 }

 return 0;
}

int readMIDI(struct midi_device *dev, struct buffer *buffer)
{
 int c, r;
 unsigned char data[MIDI_BUFFER_SIZE];

 if((r=snd_rawmidi_read(dev->input, &data, MIDI_BUFFER_SIZE))<0)
 {
  if(r==-11)
   return -1;

  printf("Can't read from MIDI device: %s\n", snd_strerror(r));
  return 4;
 }

 if(r)
 {
  printf("MIDI read  %d: ", r);
  for(c=0;c<r;c++)
   printf("%.2x ", data[c]);

  printf("\n");
 }

 if(buffer)
  copyBuffer(buffer, data, r);

 return 0;
}

int writeMIDI(struct midi_device *dev, struct buffer *buffer)
{
 int c, r;

 if(!buffer||!buffer->length)
 {
  printf("Nothing to write to MIDI device\n");
  return 4;
 }

 if(buffer->length) //data
 {
  printf("MIDI write %d: ", buffer->length);
  for(c=0;c<buffer->length;c++)
   printf("%.2x ", buffer->data[c]);

  printf("\n");
 }

 if((r=snd_rawmidi_write(dev->output, buffer->data, buffer->length))<0)
 {
  printf("Can't write to MIDI device: %s\n", snd_strerror(r));
  return 4;
 }

 return 0;
}


int main(int argc, char **argv)
{
 int c, r;
 struct buffer rolandIn, rolandOut, midiIn, midiOut;

 struct usb_device rolandTD15;

 int midiSkipOffs=0;
 char midiID[MIDI_ID_SIZE]="";
 struct midi_device midi;

 /*
  * Arguments
  */

 if(argc>1)
 {
  if(strcmp(argv[1], "--help")==0||strcmp(argv[1], "-h")==0)
  {
   printf("\n%s [midi device] [skipOffs]\n\n", argv[0]);
   printf("Optional paramaters:\n");
   printf("- midi device:  Device which should receive midi signals (e.g. 0,0)\n");
   printf("- skipOffs:     Skip midi off signals\n\n");
   return 0;
  }


  //setup midi
  strncpy(midiID, argv[1], MIDI_ID_SIZE);

  if(argc>2&&strcmp(argv[2], "skipOffs")==0)
   midiSkipOffs=1;
 }


 /*
  * Init
  */

 if((r=initSignals())!=0)
  return r;

 if(
  (r=initBuffer(&rolandIn))!=0||
  (r=initBuffer(&rolandOut))!=0||
  (r=initBuffer(&midiIn))!=0||
  (r=initBuffer(&midiOut))!=0
 )
  return r;

 if((r=initRolandTD15(&rolandTD15))!=0)
  return r;

 if(strcmp(midiID, "")!=0)
  if((r=initMIDI(&midi, midiID))!=0)
   return r;



 /*
  * Main
  */


 rolandOut.length=4; //out is always 4 bytes
 while(!quit)
 {
  //read from MIDI
  if(strcmp(midiID, "")!=0)
  {
   if((r=readMIDI(&midi, &midiIn))>0)
   {
    quit=1;
    break;
   }

   if(r==0)
   {
    //convert midi signal
    for(c=0;c<midiIn.length;c++)
    {
     switch(midiIn.data[c+0])
     {
      case 0xB0: //empty..?
      {
       c+=2; //skip next 2 bytes
       continue;
      }

      case 0x80: //stop notes
      case 0x90: //start notes
      {
       if(midiIn.data[c+0]==0x80) //stop
        rolandOut.data[0]=8; //type
       else //start
        rolandOut.data[0]=9; //type

       rolandOut.data[1]=midiIn.data[c+0]; //type
       rolandOut.data[2]=midiIn.data[c+1]; //instrument
       rolandOut.data[3]=midiIn.data[c+2]; //volume

       c+=2; //skip 2 in data
       break;
      }
      default: //continuation of previous note
      {
       rolandOut.data[2]=midiIn.data[c+0]; //instrument
       rolandOut.data[3]=midiIn.data[c+1]; //volume

       c++; //skip 1 in data
       break;
      }
     }

     //write data to roland
     if((r=writeRolandTD15(&rolandTD15, &rolandOut))!=0)
     {
      quit=1;
      break;
     }
    }
   }
  }

  //Read from roland
  if((r=readRolandTD15(&rolandTD15, &rolandIn))!=0)
  {
   quit=1;
   break;
  }

  if(rolandIn.length==4&&strcmp(midiID, "")!=0) //midi, send to midi device
  {
   if(rolandIn.data[0]==0) //skip empty data
    continue;

   if(midiSkipOffs&&rolandIn.data[0]!=9) //off signals
    continue;

   //write data to midi
   if((r=writeMIDI(&midi, &rolandIn))!=0)
   {
    quit=1;
    break;
   }
  }
 }


 /*
  * Shutdown
  */

 if(strcmp(midiID, "")!=0)
 {
  if((r=deinitMIDI(&midi))!=0)
  {
   deinitRolandTD15(&rolandTD15);
   deinitBuffer(&rolandIn);
   deinitBuffer(&rolandOut);
   deinitBuffer(&midiIn);
   deinitBuffer(&midiOut);
   return r;
  }
 }

 if((r=deinitRolandTD15(&rolandTD15))!=0)
 {
  deinitBuffer(&rolandIn);
  deinitBuffer(&rolandOut);
  deinitBuffer(&midiIn);
  deinitBuffer(&midiOut);
  return r;
 }

 if(
  (r=deinitBuffer(&rolandIn))!=0||
  (r=deinitBuffer(&rolandOut))!=0||
  (r=deinitBuffer(&midiIn))!=0||
  (r=deinitBuffer(&midiOut))!=0
 )

 if((r=deinitBuffer(&rolandIn))!=0)
  return r;

 return 0;
}

/*
 INIT
 ffff88005e334300 458923745 S Ci:2:013:0 s 80 06 0100 0000 0012 18 <
 ffff88005e334300 458923925 C Ci:2:013:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff88005e334300 458923984 S Ci:2:013:0 s 80 06 0200 0000 0009 9 <
 ffff88005e334300 458924170 C Ci:2:013:0 0 9 = 0902a700 030100c0 00
 ffff88005e334300 458924192 S Ci:2:013:0 s 80 06 0200 0000 00a7 167 <
 ffff88005e334300 458924420 C Ci:2:013:0 0 167 = 0902a700 030100c0 00090400 0000ff02 02000624 f1010000 09040001 01ff0202
 ffff8801fc4ff000 459221430 S Ci:2:013:0 s 80 06 0100 0000 0040 64 <
 ffff8801fc4ff000 459221550 C Ci:2:013:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff8801fc4ff000 459415239 S Ci:2:013:0 s 80 06 0100 0000 0012 18 <
 ffff8801fc4ff000 459415696 C Ci:2:013:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff8801fc4ff000 459418233 S Ci:2:013:0 s 80 06 0200 0000 00ff 255 <
 ffff8801fc4ff000 459418659 C Ci:2:013:0 0 167 = 0902a700 030100c0 00090400 0000ff02 02000624 f1010000 09040001 01ff0202
 ffff8801fc4ff000 459420199 S Ci:2:013:0 s 80 06 0300 0000 00ff 255 <
 ffff8801fc4ff000 459420653 C Ci:2:013:0 0 4 = 04030904
 ffff8801fc4ff000 459422237 S Ci:2:013:0 s 80 06 0302 0409 00ff 255 <
 ffff8801fc4ff000 459422422 C Ci:2:013:0 0 12 = 0c035400 44002d00 31003500
 ffff8801fc4ff000 459431481 S Ci:2:013:0 s 80 06 0100 0000 0012 18 <
 ffff8801fc4ff000 459431925 C Ci:2:013:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff8801fc4ff000 459434248 S Ci:2:013:0 s 80 06 0200 0000 0050 80 <
 ffff8801fc4ff000 459434422 C Ci:2:013:0 0 80 = 0902a700 030100c0 00090400 0000ff02 02000624 f1010000 09040001 01ff0202
 ffff8801fc4ff000 459437275 S Ci:2:013:0 s 80 06 0200 0000 00a7 167 <
 ffff8801fc4ff000 459437545 C Ci:2:013:0 0 167 = 0902a700 030100c0 00090400 0000ff02 02000624 f1010000 09040001 01ff0202
 ffff8801fc4ff000 459442609 S Co:2:013:0 s 01 0b 0000 0000 0000 0
 ffff8801fc4ff000 459442795 C Co:2:013:0 0 0
 ffff8801fc4ff000 459444406 S Co:2:013:0 s 01 0b 0000 0001 0000 0
 ffff8801fc4ff000 459444546 C Co:2:013:0 0 0
 ffff8801fc4ff000 459446220 S Co:2:013:0 s 01 0b 0000 0002 0000 0
 ffff8801fc4ff000 459446645 C Co:2:013:0 0 0
 ffff880171da1780 459499280 S Co:2:013:0 s 01 0b 0000 0002 0000 0
 ffff880171da1780 459499422 C Co:2:013:0 0 0
 ffff880171da1780 459501356 S Co:2:013:0 s 01 0b 0000 0000 0000 0
 ffff880171da1780 459501547 C Co:2:013:0 0 0
 ffff8800720ca000 459503515 S Co:2:013:0 s 01 0b 0000 0001 0000 0
 ffff8800720ca000 459503941 C Co:2:013:0 0 0
 ffff8800720ca000 459505607 S Bi:2:013:2 -115 512 <


 DEINIT
 ffff8800720ca000 459505607 S Bi:2:013:2 -115 512 <
 ffff8800720ca000 550499250 C Bi:2:013:2 -2 0
 ffff88020be5b9c0 550499458 S Co:2:001:0 s 23 03 0004 0004 0000 0
 ffff88020be5b9c0 550499467 C Co:2:001:0 0 0
 ffff88020be5b9c0 550550063 S Ci:2:001:0 s a3 00 0000 0004 0004 4 <
 ffff88020be5b9c0 550550299 C Ci:2:001:0 0 4 = 03051000
 ffff88020be5b9c0 550601153 S Co:2:001:0 s 23 01 0014 0004 0000 0
 ffff88020be5b9c0 550601163 C Co:2:001:0 0 0
 ffff88020be5b9c0 550601179 S Ci:2:000:0 s 80 06 0100 0000 0040 64 <
 ffff88020be5b9c0 550601256 C Ci:2:000:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff8801fc40bb40 550601276 S Co:2:001:0 s 23 03 0004 0004 0000 0
 ffff8801fc40bb40 550601282 C Co:2:001:0 0 0
 ffff8801fc40bb40 550652134 S Ci:2:001:0 s a3 00 0000 0004 0004 4 <
 ffff8801fc40bb40 550652374 C Ci:2:001:0 0 4 = 03051000
 ffff8801fc40bb40 550702980 S Co:2:001:0 s 23 01 0014 0004 0000 0
 ffff8801fc40bb40 550702990 C Co:2:001:0 0 0
 ffff8801fc40bb40 550702994 S Co:2:000:0 s 00 05 000d 0000 0000 0
 ffff8801fc40bb40 550703130 C Co:2:000:0 0 0
 ffff8801fc40bb40 550714892 S Ci:2:013:0 s 80 06 0100 0000 0012 18 <
 ffff8801fc40bb40 550715484 C Ci:2:013:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff8801fc40bc00 550715496 S Ci:2:013:0 s 80 06 0200 0000 00a7 167 <
 ffff8801fc40bc00 550715750 C Ci:2:013:0 0 167 = 0902a700 030100c0 00090400 0000ff02 02000624 f1010000 09040001 01ff0202
 ffff8801fc40bb40 550715758 S Co:2:013:0 s 00 09 0001 0000 0000 0
 ffff8801fc40bb40 550715878 C Co:2:013:0 0 0
*/
